@@ -18,6 +18,7 @@ import jax
 from jax import test_util as jtu
 import numpy as np
 import os
+import unittest
 import sys
 
 from jax.experimental.jax2tf.tests import tf_test_util
@@ -43,11 +44,12 @@ from jax.config import config
 
 config.parse_flags_with_absl()
 
-
 class StaxTest(tf_test_util.JaxToTfTestCase):
 
   @jtu.skip_on_flag("jax_skip_slow_tests", True)
   def test_res_net(self):
+    if config.FLAGS.jax_enable_x64:
+      raise unittest.SkipTest("ResNet test fails on JAX when X64 is enabled")
     key = jax.random.PRNGKey(0)
     shape = (224, 224, 3, 1)
     init_fn, apply_fn = resnet50.ResNet50(1000)
