@@ -16,6 +16,7 @@ Parallelization primitives.
 """
 
 import collections
+import warnings
 
 import numpy as np
 
@@ -243,9 +244,7 @@ def pswapaxes(x, axis_name, axis):
       ``axis_name``.
 
   Returns:
-    Array(s) with shape ``np.insert(np.delete(x.shape, axis), axis, axis_size)``
-    where ``axis_size`` is the size of the mapped axis named ``axis_name`` in
-    the input ``x``.
+    Array(s) with the same shape as ``x``.
   """
   return all_to_all(x, axis_name, axis, axis)
 
@@ -492,6 +491,7 @@ pmax_p = core.Primitive('pmax')
 pmax_p.def_abstract_eval(lambda x, **params: raise_to_shaped(x))
 xla.parallel_translations[pmax_p] = \
     partial(_allreduce_translation_rule, lax.max_p)
+pxla.multi_host_supported_collectives.add(pmax_p)
 batching.split_axis_rules[pmax_p] = partial(_split_axis_comm_assoc, pmax_p)
 batching.primitive_batchers[pmax_p] = partial(_collective_batcher, pmax_p)
 batching.collective_rules[pmax_p] = \
@@ -505,6 +505,7 @@ pmin_p = core.Primitive('pmin')
 pmin_p.def_abstract_eval(lambda x, **params: raise_to_shaped(x))
 xla.parallel_translations[pmin_p] = \
     partial(_allreduce_translation_rule, lax.min_p)
+pxla.multi_host_supported_collectives.add(pmin_p)
 batching.split_axis_rules[pmin_p] = partial(_split_axis_comm_assoc, pmin_p)
 batching.primitive_batchers[pmin_p] = partial(_collective_batcher, pmin_p)
 batching.collective_rules[pmin_p] = \
@@ -551,45 +552,127 @@ batching.primitive_batchers[ppermute_p] = partial(_collective_batcher, ppermute_
 batching.collective_rules[ppermute_p] = _ppermute_batcher
 
 
+def _moveaxis(src, dst, x):
+  perm = [i for i in range(x.ndim) if i != src]
+  perm.insert(dst, src)
+  return lax.transpose(x, perm)
+
+def _all_to_all_via_all_gather(x, *, axis_name, split_axis, concat_axis):
+  global_full = all_gather(x, axis_name)
+  idx = axis_index(axis_name)
+  local_slice = lax.dynamic_index_in_dim(global_full, idx, split_axis + 1, keepdims=False)
+  return _moveaxis(0, concat_axis, local_slice)
+
 def _all_to_all_translation_rule(c, x, *, split_axis, concat_axis, axis_name,
                                  axis_env, platform):
   # Workaround for AllToAll not being implemented on CPU.
   replica_groups = _replica_groups(axis_env, axis_name, None)
   if len(replica_groups[0]) == 1:
     return x
+  elif platform != 'tpu':
+    warnings.warn("all_to_all (and pswapaxes) are only implemented properly for TPUs. All other "
+                  "backends emulate it using a very slow and memory intensive algorithm, so expect "
+                  "significant slowdowns.")
+    lowering = xla.lower_fun(_all_to_all_via_all_gather, multiple_results=False, parallel=True)
+    return lowering(c, x,
+                    split_axis=split_axis, concat_axis=concat_axis, axis_name=axis_name,
+                    axis_env=axis_env, platform=platform)
   else:
     split_count = len(replica_groups[0])
     if not all(split_count == len(g) for g in replica_groups):
       raise ValueError('Replica groups must be equally sized')
     replica_groups_protos = xc.make_replica_groups(replica_groups)
-    return xops.AllToAll(x, split_axis, concat_axis, split_count,
-                         replica_groups_protos)
-
-def _all_to_all_split_axis_rule(vals, which_mapped, split_axis, concat_axis,
-                                axis_name):
-  assert tuple(which_mapped) == (True,)
-  x, = vals
-  # perform the communication to swap the hardware-mapped axes
-  stacked = all_to_all_p.bind(x, split_axis=split_axis + 1, concat_axis=0,
-                              axis_name=axis_name)
-  # transpose the newly mapped axis to the front, newly unmapped to concat_axis
-  out = _moveaxis(split_axis + 1, 0, stacked)
-  out = _moveaxis(1, concat_axis + 1, out)
-  return out, True
+    if concat_axis == split_axis:
+      return xops.AllToAll(x, split_axis, concat_axis, split_count,
+                           replica_groups_protos)
+    else:
+      if concat_axis < split_axis:
+        split_axis += 1
+      elif split_axis < concat_axis:
+        concat_axis += 1
+      x = xla.lower_fun(partial(lax.expand_dims, dimensions=(concat_axis,)), multiple_results=False)(c, x)
+      x = xops.AllToAll(x, split_axis, concat_axis, split_count, replica_groups_protos)
+      x = xla.lower_fun(partial(lax.squeeze, dimensions=(split_axis,)), multiple_results=False)(c, x)
+      return x
 
 def _all_to_all_transpose_rule(cts, axis_name, split_axis, concat_axis):
   return (all_to_all(cts, axis_name=axis_name, split_axis=concat_axis, concat_axis=split_axis),)
 
-def _moveaxis(src, dst, x):
-  perm = [i for i in range(x.ndim) if i != src]
-  perm.insert(dst, src)
-  return lax.transpose(x, perm)
+def _all_to_all_batcher(vals_in, dims_in, *, axis_name, split_axis, concat_axis):
+  x, = vals_in
+  d, = dims_in
+  if d <= split_axis:
+    split_axis += 1
+  if d <= concat_axis:
+    concat_axis += 1
+  # Note: At this point split_axis and concat_axis are adjusted to the extra
+  #       dimension and we have d != split_axis and d != concat_axis.
+  if split_axis < d < concat_axis:
+    d -= 1
+  elif concat_axis < d < split_axis:
+    d += 1
+  result = all_to_all_p.bind(x, axis_name=axis_name, split_axis=split_axis, concat_axis=concat_axis)
+  return result, d
+
+def _all_to_all_batched_collective(vals_in, dims_in, axis_size, axis_name, split_axis, concat_axis):
+  x, = vals_in
+  d, = dims_in
+  split_axis_adj = split_axis + (1 if d <= split_axis else 0)
+  concat_axis_adj = concat_axis + (1 if split_axis_adj <= concat_axis else 0)
+  if d < split_axis_adj < concat_axis_adj:
+    split_axis_adj -= 1
+  elif concat_axis_adj < split_axis_adj < d:
+    split_axis_adj += 1
+  return [_moveaxis(d, concat_axis_adj, x)], [split_axis_adj]
+
+def _all_to_all_split_axis_rule(split_name, vals, params):
+  concat_axis = params['concat_axis']
+  split_axis = params['split_axis']
+  axis_names = params['axis_name']
+  assert isinstance(axis_names, tuple)
+  x, = vals
+
+  split_pos = list(axis_names).index(split_name)
+  before_axes = axis_names[:split_pos]
+  after_axes = axis_names[split_pos+1:]
+
+  # Flatten the split_dim
+  split_name_size = psum(1, split_name)
+  before_size = psum(1, before_axes)
+  after_size = psum(1, after_axes)
+  unroll_shape = list(x.shape)
+  unroll_shape[split_axis:split_axis+1] = [before_size, split_name_size, after_size]
+  unroll_x = lax.reshape(x, unroll_shape)
+
+  if before_axes:
+    out_before = all_to_all(unroll_x, before_axes, split_axis, concat_axis=0)
+  else:
+    out_before = _moveaxis(split_axis, 0, unroll_x)
+  out_split = all_to_all(out_before, split_name, split_axis + 1, concat_axis=1)
+  if after_axes:
+    out_after = all_to_all(out_split, after_axes, split_axis + 2, concat_axis=2)
+  else:
+    out_after = _moveaxis(split_axis + 2, 2, out_split)
+
+  # Flatten the concat axes and move them to the right position
+  y = out_after.reshape((np.prod(out_after.shape[:3]), *out_after.shape[3:]))
+  return _moveaxis(0, concat_axis, y)
+
+def _all_to_all_abstract_eval(x, axis_name, split_axis, concat_axis):
+  input_aval = raise_to_shaped(x)
+  shape = list(input_aval.shape)
+  size = shape.pop(split_axis)
+  shape.insert(concat_axis, size)
+  return ShapedArray(tuple(shape), input_aval.dtype, weak_type=False)
 
 all_to_all_p = core.Primitive('all_to_all')
-all_to_all_p.def_abstract_eval(lambda x, **params: raise_to_shaped(x))
+all_to_all_p.def_abstract_eval(_all_to_all_abstract_eval)
 xla.parallel_translations[all_to_all_p] = _all_to_all_translation_rule
 ad.deflinear(all_to_all_p, _all_to_all_transpose_rule)
 pxla.multi_host_supported_collectives.add(all_to_all_p)
+batching.primitive_batchers[all_to_all_p] = _all_to_all_batcher
+batching.collective_rules[all_to_all_p] = _all_to_all_batched_collective
+batching.split_axis_rules[all_to_all_p] = _all_to_all_split_axis_rule
 
 
 def _expand(dim, size, index, x):
