@@ -13,7 +13,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-
 import numpy as np
 
 from jax.numpy import lax_numpy as jnp
@@ -47,9 +46,9 @@ def cholesky(x, symmetrize_input=True):
     x = symmetrize(x)
   return jnp.tril(cholesky_p.bind(x))
 
-def eig(x):
-  w, vl, vr = eig_p.bind(x)
-  return w, vl, vr
+def eig(x, compute_left_eigenvectors=True, compute_right_eigenvectors=True):
+  return eig_p.bind(x, compute_left_eigenvectors=compute_left_eigenvectors,
+                    compute_right_eigenvectors=compute_right_eigenvectors)
 
 def eigh(x, lower=True, symmetrize_input=True):
   if symmetrize_input:
@@ -58,18 +57,26 @@ def eigh(x, lower=True, symmetrize_input=True):
   return v, w
 
 def lu(x):
-  lu, pivots = lu_p.bind(x)
-  return lu, pivots
+  lu, pivots, permutation = lu_p.bind(x)
+  return lu, pivots, permutation
 
 def qr(x, full_matrices=True):
   q, r = qr_p.bind(x, full_matrices=full_matrices)
   return q, r
 
 def svd(x, full_matrices=True, compute_uv=True):
-  s, u, v = svd_p.bind(x, full_matrices=full_matrices, compute_uv=compute_uv)
+  """Singular value decomposition.
+
+  Returns the singular values if compute_uv is False, otherwise returns a triple
+  containing the left singular vectors, the singular values and the adjoint of
+  the right singular vectors.
+  """
+  result = svd_p.bind(x, full_matrices=full_matrices, compute_uv=compute_uv)
   if compute_uv:
+    s, u, v = result
     return u, s, v
   else:
+    s, = result
     return s
 
 def triangular_solve(a, b, left_side=False, lower=False, transpose_a=False,
@@ -158,14 +165,19 @@ xla.backend_specific_translations['gpu'][cholesky_p] = partial(
 
 # Asymmetric eigendecomposition
 
-def eig_impl(operand):
-  return xla.apply_primitive(eig_p, operand)
+def eig_impl(operand, *, compute_left_eigenvectors, compute_right_eigenvectors):
+  return (
+    xla.apply_primitive(eig_p, operand,
+                        compute_left_eigenvectors=compute_left_eigenvectors,
+                        compute_right_eigenvectors=compute_right_eigenvectors))
 
-def eig_translation_rule(c, operand):
+def eig_translation_rule(c, operand, *, compute_left_eigenvectors,
+                         compute_right_eigenvectors):
   raise NotImplementedError(
     "Nonsymmetric eigendecomposition is only implemented on the CPU backend")
 
-def eig_abstract_eval(operand):
+def eig_abstract_eval(operand, *, compute_left_eigenvectors,
+                      compute_right_eigenvectors):
   if isinstance(operand, ShapedArray):
     if operand.ndim < 2 or operand.shape[-2] != operand.shape[-1]:
       raise ValueError("Argument to nonsymmetric eigendecomposition must have "
@@ -179,28 +191,51 @@ def eig_abstract_eval(operand):
     w = ShapedArray(batch_dims + (n,), dtype)
   else:
     raise NotImplementedError
-  return w, vl, vr
+
+  output = [w]
+  if compute_left_eigenvectors:
+    output.append(vl)
+  if compute_right_eigenvectors:
+    output.append(vr)
+
+  return tuple(output)
 
 _cpu_geev = lapack.geev
 
-def eig_cpu_translation_rule(c, operand):
+def eig_cpu_translation_rule(c, operand, *, compute_left_eigenvectors,
+                             compute_right_eigenvectors):
   shape = c.get_shape(operand)
   batch_dims = shape.dimensions()[:-2]
-  w, vl, vr, info = _cpu_geev(c, operand)
+
+  w, vl, vr, info = _cpu_geev(c, operand, jobvl=compute_left_eigenvectors,
+                              jobvr=compute_right_eigenvectors)
+
   ok = xops.Eq(info, xops.ConstantLiteral(c, np.array(0, np.int32)))
   w = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1,)), w,
                            _nan_like(c, w))
-  vl = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), vl,
-                            _nan_like(c, vl))
-  vr = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), vr,
-                            _nan_like(c, vr))
-  return xops.Tuple(c, [w, vl, vr])
+  output = [w]
 
-def eig_batching_rule(batched_args, batch_dims):
+  if compute_left_eigenvectors:
+    vl = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), vl,
+                              _nan_like(c, vl))
+    output.append(vl)
+
+  if compute_right_eigenvectors:
+    vr = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), vr,
+                              _nan_like(c, vr))
+    output.append(vr)
+
+  return xops.Tuple(c, output)
+
+def eig_batching_rule(batched_args, batch_dims, *, compute_left_eigenvectors,
+                      compute_right_eigenvectors):
   x, = batched_args
   bd, = batch_dims
   x = batching.moveaxis(x, bd, 0)
-  return eig_p.bind(x), (0, 0, 0)
+
+  return (eig_p.bind(x, compute_left_eigenvectors=compute_left_eigenvectors,
+                     compute_right_eigenvectors=compute_right_eigenvectors),
+          (0,) * (1 + compute_left_eigenvectors + compute_right_eigenvectors))
 
 eig_p = Primitive('eig')
 eig_p.multiple_results = True
@@ -526,13 +561,15 @@ def _lu_blocked(a, block_size=128):
   m, n = a.shape
   r = min(m, n)
   pivot = jnp.zeros((r,), dtype=jnp.int32)
+  perm = jnp.arange(m, dtype=jnp.int32)
   for k in range(0, r, block_size):
     b = min(r - k, block_size)
-    block_pivot, perm, lu_block = _lu_unblocked(a[k:, k:k+b])
+    block_pivot, block_perm, lu_block = _lu_unblocked(a[k:, k:k+b])
 
-    a = ops.index_update(a, ops.index[k:, :], a[perm + k, :])
-    a = ops.index_update(a, ops.index[k:, k:k+b], lu_block)
     pivot = ops.index_update(pivot, ops.index[k:k+b], block_pivot + k)
+    perm = ops.index_update(perm, ops.index[k:], perm[block_perm + k])
+    a = ops.index_update(a, ops.index[k:, :], a[block_perm + k, :])
+    a = ops.index_update(a, ops.index[k:, k:k+b], lu_block)
 
     if k + b < n:
       a = ops.index_update(
@@ -543,7 +580,7 @@ def _lu_blocked(a, block_size=128):
         a, ops.index[k+b:, k+b:],
         -lax.dot(a[k+b:, k:k+b], a[k:k+b, k+b:],
                  precision=lax.Precision.HIGHEST))
-  return pivot, a
+  return a, pivot, perm
 
 def _lu_python(x):
   """Default LU decomposition in Python, where no better version exists."""
@@ -551,16 +588,17 @@ def _lu_python(x):
   batch_dims = x.shape[:-2]
   if len(batch_dims) > 0:
     batch_size = np.prod(batch_dims, dtype=np.int64)
-    pivot, lu = api.vmap(_lu_blocked)(lax.reshape(x, (batch_size, m, n)))
-    pivot = lax.reshape(pivot, batch_dims + (min(m, n),))
+    lu, pivot, perm = api.vmap(_lu_blocked)(lax.reshape(x, (batch_size, m, n)))
     lu = lax.reshape(lu, batch_dims + (m, n))
+    pivot = lax.reshape(pivot, batch_dims + (min(m, n),))
+    perm = lax.reshape(perm, batch_dims + (m,))
   else:
-    pivot, lu = _lu_blocked(x)
-  return lu, pivot
+    lu, pivot, perm = _lu_blocked(x)
+  return lu, pivot, perm
 
 def _lu_impl(operand):
-  lu, pivot = xla.apply_primitive(lu_p, operand)
-  return lu, pivot
+  lu, pivot, perm = xla.apply_primitive(lu_p, operand)
+  return lu, pivot, perm
 
 def _lu_abstract_eval(operand):
   if isinstance(operand, ShapedArray):
@@ -571,21 +609,22 @@ def _lu_abstract_eval(operand):
     m = operand.shape[-2]
     n = operand.shape[-1]
     pivot = ShapedArray(batch_dims + (min(m, n),), jnp.int32)
+    perm = ShapedArray(batch_dims + (m,), jnp.int32)
   else:
     pivot = operand
-  return operand, pivot
+    perm = operand
+  return operand, pivot, perm
 
 def _lu_jvp_rule(primals, tangents):
   a, = primals
   a_dot, = tangents
-  lu, pivots = lu_p.bind(a)
+  lu, pivots, permutation = lu_p.bind(a)
 
   a_shape = jnp.shape(a)
   m, n = a_shape[-2:]
   dtype = lax.dtype(a)
   k = min(m, n)
 
-  permutation = lu_pivots_to_permutation(pivots, m)
   batch_dims = a_shape[:-2]
   iotas = jnp.ix_(*(lax.iota(jnp.int32, b) for b in batch_dims + (1,)))
   x = a_dot[iotas[:-1] + (permutation, slice(None))]
@@ -620,25 +659,37 @@ def _lu_jvp_rule(primals, tangents):
   l_dot = jnp.matmul(l, jnp.tril(lau, -1))
   u_dot = jnp.matmul(jnp.triu(lau), u)
   lu_dot = l_dot + u_dot
-  return (lu, pivots), (lu_dot, ad_util.Zero.from_value(pivots))
+  return (lu, pivots, permutation), (lu_dot, ad_util.Zero.from_value(pivots),
+                                     ad_util.Zero.from_value(permutation))
 
 
 def _lu_batching_rule(batched_args, batch_dims):
   x, = batched_args
   bd, = batch_dims
   x = batching.moveaxis(x, bd, 0)
-  return lu_p.bind(x), (0, 0)
+  return lu_p.bind(x), (0, 0, 0)
 
 def _lu_cpu_gpu_translation_rule(getrf_impl, c, operand):
   shape = c.get_shape(operand)
   batch_dims = shape.dimensions()[:-2]
+  m = shape.dimensions()[-2]
   lu, pivot, info = getrf_impl(c, operand)
   # Subtract 1 from the pivot to get 0-based indices.
   pivot = xops.Sub(pivot, xops.ConstantLiteral(c, np.array(1, np.int32)))
   ok = xops.Ge(info, xops.ConstantLiteral(c, np.array(0, np.int32)))
   lu = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), lu,
                             _nan_like(c, lu))
-  return xops.Tuple(c, [lu, pivot])
+  perm = xla.lower_fun(lambda x: lu_pivots_to_permutation(x, m),
+                       multiple_results=False)(c, pivot)
+  return xops.Tuple(c, [lu, pivot, perm])
+
+
+def _lu_tpu_translation_rule(c, operand):
+  if hasattr(xops, "LU"):
+    lu, pivot, perm = xops.LU(operand)
+    return xops.Tuple(c, [lu, pivot, perm])
+  else:
+    return xla.lower_fun(_lu_python, multiple_results=True)(c, operand)
 
 
 lu_p = Primitive('lu')
@@ -655,6 +706,8 @@ xla.backend_specific_translations['cpu'][lu_p] = partial(
 xla.backend_specific_translations['gpu'][lu_p] = partial(
   _lu_cpu_gpu_translation_rule, cusolver.getrf)
 
+xla.backend_specific_translations['tpu'][lu_p] = _lu_tpu_translation_rule
+
 
 # Define this outside lu_pivots_to_permutation to ensure fori_loop cache hits
 def _lu_pivots_body_fn(i, permutation_and_swaps):
@@ -687,15 +740,16 @@ def lu_pivots_to_permutation(swaps, m):
 
   permutation = lax.broadcasted_iota(jnp.int32, batch_dims + (m,),
                                      len(batch_dims))
+  if m == 0:
+    return permutation
   result, _ = lax.fori_loop(np.array(0, np.int32), np.array(k, np.int32),
                             _lu_pivots_body_fn, (permutation, swaps))
   return result
 
 
 @partial(vectorize, excluded={3}, signature='(n,n),(n),(n,k)->(n,k)')
-def _lu_solve_core(lu, pivots, b, trans):
+def _lu_solve_core(lu, permutation, b, trans):
   m = lu.shape[0]
-  permutation = lu_pivots_to_permutation(pivots, m)
   x = jnp.reshape(b, (m, -1))
   if trans == 0:
     x = x[permutation, :]
@@ -714,7 +768,7 @@ def _lu_solve_core(lu, pivots, b, trans):
 
 
 @partial(api.jit, static_argnums=(3,))
-def _lu_solve(lu, pivots, b, trans):
+def _lu_solve(lu, permutation, b, trans):
   if len(lu.shape) < 2 or lu.shape[-1] != lu.shape[-2]:
     raise ValueError("last two dimensions of LU decomposition must be equal, "
                      "got shape {}".format(lu.shape))
@@ -739,13 +793,13 @@ def _lu_solve(lu, pivots, b, trans):
                        "matrix (shape {}) and second to last axis of b array "
                        "(shape {}) must match"
                        .format(lu.shape, b.shape))
-  x = _lu_solve_core(lu, pivots, b, trans)
+  x = _lu_solve_core(lu, permutation, b, trans)
   return x[..., 0] if rhs_vector else x
 
 
-def lu_solve(lu, pivots, b, trans=0):
+def lu_solve(lu, permutation, b, trans=0):
   """LU solve with broadcasting."""
-  return _lu_solve(lu, pivots, b, trans)
+  return _lu_solve(lu, permutation, b, trans)
 
 
 # QR decomposition
@@ -847,9 +901,8 @@ xla.backend_specific_translations['gpu'][qr_p] = partial(
 # Singular value decomposition
 
 def svd_impl(operand, full_matrices, compute_uv):
-  s, u, vt = xla.apply_primitive(svd_p, operand, full_matrices=full_matrices,
-                                 compute_uv=compute_uv)
-  return s, u, vt
+  return xla.apply_primitive(svd_p, operand, full_matrices=full_matrices,
+                             compute_uv=compute_uv)
 
 def svd_translation_rule(c, operand, full_matrices, compute_uv):
   shape = c.get_shape(operand).dimensions()
@@ -861,7 +914,12 @@ def svd_translation_rule(c, operand, full_matrices, compute_uv):
   if not full_matrices and m != n:
     u = xops.SliceInDim(u, 0, min(m, n), stride=1, dimno=len(shape) - 1)
     vt = xops.SliceInDim(vt, 0, min(m, n), stride=1, dimno=len(shape) - 2)
-  return xops.Tuple(c, [s, u, vt])
+
+  if not compute_uv:
+    return xops.Tuple(c, [s])
+  else:
+    return xops.Tuple(c, [s, u, vt])
+
 
 def svd_abstract_eval(operand, full_matrices, compute_uv):
   if isinstance(operand, ShapedArray):
@@ -872,11 +930,14 @@ def svd_abstract_eval(operand, full_matrices, compute_uv):
     m = operand.shape[-2]
     n = operand.shape[-1]
     s = ShapedArray(batch_dims + (min(m, n),), lax.lax._complex_basetype(operand.dtype))
-    u = ShapedArray(batch_dims + (m, m if full_matrices else min(m, n)), operand.dtype)
-    vt = ShapedArray(batch_dims + (n if full_matrices else min(m, n), n), operand.dtype)
+    if compute_uv:
+      u = ShapedArray(batch_dims + (m, m if full_matrices else min(m, n)), operand.dtype)
+      vt = ShapedArray(batch_dims + (n if full_matrices else min(m, n), n), operand.dtype)
+      return s, u, vt
+    else:
+      return s,
   else:
     raise NotImplementedError
-  return s, u, vt
 
 def svd_jvp_rule(primals, tangents, full_matrices, compute_uv):
   A, = primals
@@ -893,6 +954,10 @@ def svd_jvp_rule(primals, tangents, full_matrices, compute_uv):
   s_dim = s[..., None, :]
   dS = jnp.matmul(jnp.matmul(Ut, dA), V)
   ds = jnp.real(jnp.diagonal(dS, 0, -2, -1))
+
+  if not compute_uv:
+    return (s,), (ds,)
+
   F = 1 / (jnp.square(s_dim) - jnp.square(_T(s_dim)) + jnp.eye(k, dtype=A.dtype))
   F = F - jnp.eye(k, dtype=A.dtype)
   dSS = s_dim * dS
@@ -917,18 +982,28 @@ def _svd_cpu_gpu_translation_rule(gesvd_impl, c, operand, full_matrices, compute
   ok = xops.Eq(info, xops.ConstantLiteral(c, np.array(0, np.int32)))
   s = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1,)), s,
                            _nan_like(c, s))
-  u = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), u,
-                           _nan_like(c, u))
-  vt = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), vt,
-                            _nan_like(c, vt))
-  return xops.Tuple(c, [s, u, vt])
+
+  result = [s]
+
+  if compute_uv:
+    u = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), u,
+                             _nan_like(c, u))
+    vt = _broadcasting_select(c, xops.Reshape(ok, batch_dims + (1, 1)), vt,
+                              _nan_like(c, vt))
+    result += [u, vt]
+
+  return xops.Tuple(c, result)
 
 def svd_batching_rule(batched_args, batch_dims, full_matrices, compute_uv):
   x, = batched_args
   bd, = batch_dims
   x = batching.moveaxis(x, bd, 0)
   outs = svd_p.bind(x, full_matrices=full_matrices, compute_uv=compute_uv)
-  return outs, (0, 0, 0)
+
+  if compute_uv:
+    return outs, (0, 0, 0)
+  else:
+    return outs, (0,)
 
 svd_p = Primitive('svd')
 svd_p.multiple_results = True
