@@ -105,23 +105,25 @@ def inv(a, overwrite_a=False, check_finite=True):
 def lu_factor(a, overwrite_a=False, check_finite=True):
   del overwrite_a, check_finite
   a = np_linalg._promote_arg_dtypes(jnp.asarray(a))
-  return lax_linalg.lu(a)
+  lu, pivots, _ = lax_linalg.lu(a)
+  return lu, pivots
 
 
 @_wraps(scipy.linalg.lu_solve)
 def lu_solve(lu_and_piv, b, trans=0, overwrite_b=False, check_finite=True):
   del overwrite_b, check_finite
   lu, pivots = lu_and_piv
-  return lax_linalg.lu_solve(lu, pivots, b, trans)
+  m, n = lu.shape[-2:]
+  perm = lax_linalg.lu_pivots_to_permutation(pivots, m)
+  return lax_linalg.lu_solve(lu, perm, b, trans)
 
 
 @partial(jit, static_argnums=(1,))
 def _lu(a, permute_l):
   a = np_linalg._promote_arg_dtypes(jnp.asarray(a))
-  lu, pivots = lax_linalg.lu(a)
+  lu, pivots, permutation = lax_linalg.lu(a)
   dtype = lax.dtype(a)
   m, n = jnp.shape(a)
-  permutation = lax_linalg.lu_pivots_to_permutation(pivots, m)
   p = jnp.real(jnp.array(permutation == jnp.arange(m)[:, None], dtype=dtype))
   k = min(m, n)
   l = jnp.tril(lu, -1)[:, :k] + jnp.eye(m, k, dtype=dtype)
@@ -236,18 +238,36 @@ def triu(m, k=0):
 _expm_description = textwrap.dedent("""
 In addition to the original NumPy argument(s) listed below,
 also supports the optional boolean argument ``upper_triangular``
-to specify whether the ``A`` matrix is upper triangular.
+to specify whether the ``A`` matrix is upper triangular, and the optional
+argument ``max_squarings`` to specify the max number of squarings allowed
+in the scaling-and-squaring approximation method. Return nan if the actual
+number of squarings required is more than ``max_squarings``.
+
+The number of required squarings = max(0, ceil(log2(norm(A)) - c)
+where norm() denotes the L1 norm, and
+  c=2.42 for float64 or complex128,
+  c=1.97 for float32 or complex64
 """)
 
 @_wraps(scipy.linalg.expm, lax_description=_expm_description)
-def expm(A, *, upper_triangular=False):
-  return _expm(A, upper_triangular)
+def expm(A, *, upper_triangular=False, max_squarings=16):
+  return _expm(A, upper_triangular, max_squarings)
 
-@partial(api.custom_jvp, nondiff_argnums=(1,))
-def _expm(A, upper_triangular):
+@partial(jit, static_argnums=(1, 2))
+def _expm(A, upper_triangular, max_squarings):
   P, Q, n_squarings = _calc_P_Q(A)
-  R = _solve_P_Q(P, Q, upper_triangular)
-  R = _squaring(R, n_squarings)
+
+  def _nan(args):
+    A, *_ = args
+    return jnp.full_like(A, jnp.nan)
+
+  def _compute(args):
+    A, P, Q = args
+    R = _solve_P_Q(P, Q, upper_triangular)
+    R = _squaring(R, n_squarings)
+    return R
+
+  R = lax.cond(n_squarings > max_squarings, _nan, _compute, (A, P, Q))
   return R
 
 @jit
@@ -268,8 +288,8 @@ def _calc_P_Q(A):
    U13, V13 = _pade13(A)
    conds=jnp.array([1.495585217958292e-002, 2.539398330063230e-001,
                     9.504178996162932e-001, 2.097847961257068e+000])
-   U = jnp.select((maxnorm<conds), (U3, U5, U7, U9), U13)
-   V = jnp.select((maxnorm<conds), (V3, V5, V7, V9), V13)
+   U = jnp.select((A_L1<conds), (U3, U5, U7, U9), U13)
+   V = jnp.select((A_L1<conds), (V3, V5, V7, V9), V13)
   elif A.dtype == 'float32' or A.dtype == 'complex64':
     U3,V3 = _pade3(A)
     U5,V5 = _pade5(A)
@@ -278,8 +298,8 @@ def _calc_P_Q(A):
     A = A / 2**n_squarings
     U7,V7 = _pade7(A)
     conds=jnp.array([4.258730016922831e-001, 1.880152677804762e+000])
-    U = jnp.select((maxnorm<conds), (U3, U5), U7)
-    V = jnp.select((maxnorm<conds), (V3, V5), V7)
+    U = jnp.select((A_L1<conds), (U3, U5), U7)
+    V = jnp.select((A_L1<conds), (V3, V5), V7)
   else:
     raise TypeError("A.dtype={} is not supported.".format(A.dtype))
   P = U + V  # p_m(A) : numerator
@@ -298,11 +318,17 @@ def _precise_dot(A, B):
 @jit
 def _squaring(R, n_squarings):
   # squaring step to undo scaling
-  def my_body_fun(i,R):
-    return _precise_dot(R, R)
-  lower = jnp.zeros(1, dtype=n_squarings.dtype)
-  R = lax.fori_loop(lower[0], n_squarings, my_body_fun, R)
-  return R
+  def _squaring_precise(x):
+    return _precise_dot(x, x)
+
+  def _identity(x):
+    return x
+
+  def _scan_f(c, i):
+    return lax.cond(i < n_squarings, _squaring_precise, _identity, c), None
+  res, _ = lax.scan(_scan_f, R, jnp.arange(16))
+
+  return res
 
 def _pade3(A):
   b = (120., 60., 12., 1.)
@@ -378,7 +404,8 @@ def _expm_frechet(A, E, method=None, compute_expm=True):
   if method is None:
     method = 'SPS'
   if method == 'SPS':
-    expm_A, expm_frechet_AE = expm_frechet_algo_64(A, E)
+    bound_fun = partial(expm, upper_triangular=False, max_squarings=16)
+    expm_A, expm_frechet_AE = api.jvp(bound_fun, (A,), (E,))
   else:
     raise ValueError('only method=\'SPS\' is supported')
   if compute_expm:
@@ -386,188 +413,6 @@ def _expm_frechet(A, E, method=None, compute_expm=True):
   else:
     return expm_frechet_AE
 
-"""
-Maximal values ell_m of ||2**-s A|| such that the backward error bound
-does not exceed 2**-53.
-"""
-ell_table_61 = (
-  None,
-  # 1
-  2.11e-8,
-  3.56e-4,
-  1.08e-2,
-  6.49e-2,
-  2.00e-1,
-  4.37e-1,
-  7.83e-1,
-  1.23e0,
-  1.78e0,
-  2.42e0,
-  # 11
-  3.13e0,
-  3.90e0,
-  4.74e0,
-  5.63e0,
-  6.56e0,
-  7.52e0,
-  8.53e0,
-  9.56e0,
-  1.06e1,
-  1.17e1,
-)
-
-@jit
-def expm_frechet_algo_64(A, E):
-  ident = jnp.eye(*A.shape, dtype=A.dtype)
-  A_norm_1 = np_linalg.norm(A, 1)
-  """
-  Subset of the Maximal values ell_m of ||2**-s A||
-  such that the backward error bound does not exceed 2**-53.
-  """
-  args = (A, E, ident, A_norm_1)
-  U3579, V3579, Lu3579, Lv3579, s3579 = lax.cond(A_norm_1 <= ell_table_61[3],
-     args, lambda args: _diff_pade3(args),
-     args, lambda args: lax.cond(A_norm_1 <= ell_table_61[5],
-         args, lambda args: _diff_pade5(args),
-         args, lambda args: lax.cond(A_norm_1 <= ell_table_61[7],
-              args, lambda args: _diff_pade7(args),
-              args, lambda args: _diff_pade9(args))))
-  U13, V13, Lu13, Lv13, s13 = _diff_pade13(args)
-
-  # Must be of minimum length 2 for np.select to be used
-  ell_table_61_local99 = jnp.array([ell_table_61[9], ell_table_61[9]])
-  U = jnp.select((A_norm_1<=ell_table_61_local99), (U3579, U3579), U13)
-  V = jnp.select((A_norm_1<=ell_table_61_local99), (V3579, V3579), V13)
-  Lu = jnp.select((A_norm_1<=ell_table_61_local99), (Lu3579, Lu3579), Lu13)
-  Lv = jnp.select((A_norm_1<=ell_table_61_local99), (Lv3579, Lv3579), Lv13)
-  s = jnp.select((A_norm_1<=ell_table_61_local99), (s3579, s3579), s13)
-
-  lu_piv = lu_factor(-U + V)
-  R = lu_solve(lu_piv, U + V)
-  L = lu_solve(lu_piv, Lu + Lv + _precise_dot((Lu - Lv), R))
-  # squaring
-  def my_body_fun(i,my_arg):
-    R, L = my_arg
-    L = _precise_dot(R, L) + _precise_dot(L, R)
-    R = _precise_dot(R, R)
-    return R, L
-  lower = jnp.zeros(1, dtype=s.dtype)
-  R, L = lax.fori_loop(lower[0], s, my_body_fun, (R, L))
-  return R, L
-
-"""
-# The b vectors and U and V are those from
-# scipy.sparse.linalg.matfuncs.py.
-# M, Lu, Lv follow (6.11), (6.12), (6.13), (3.3)
-"""
-@jit
-def _diff_pade3(args):
-  A,E,ident,_ = args
-  s = 0
-  b = (120., 60., 12., 1.)
-  A2 = A.dot(A)
-  M2 = _precise_dot(A, E) + _precise_dot(E, A)
-  U = _precise_dot(A, b[3]*A2 + b[1]*ident)
-  V = b[2]*A2 + b[0]*ident
-  Lu = _precise_dot(A, b[3]*M2) + _precise_dot(E, b[3]*A2 + b[1]*ident)
-  Lv = b[2]*M2
-  return U, V, Lu, Lv, s
-
-@jit
-def _diff_pade5(args):
-  A,E,ident,_ = args
-  s = 0
-  b = (30240., 15120., 3360., 420., 30., 1.)
-  A2 = _precise_dot(A, A)
-  M2 = _precise_dot(A, E) + _precise_dot(E, A)
-  A4 = _precise_dot(A2, A2)
-  M4 = _precise_dot(A2, M2) + _precise_dot(M2, A2)
-  U = _precise_dot(A, b[5]*A4 + b[3]*A2 + b[1]*ident)
-  V = b[4]*A4 + b[2]*A2 + b[0]*ident
-  Lu = (_precise_dot(A, b[5]*M4 + b[3]*M2) +
-          _precise_dot(E, b[5]*A4 + b[3]*A2 + b[1]*ident))
-  Lv = b[4]*M4 + b[2]*M2
-  return U, V, Lu, Lv, s
-
-@jit
-def _diff_pade7(args):
-  A, E, ident, _ = args
-  s = 0
-  b = (17297280., 8648640., 1995840., 277200., 25200., 1512., 56., 1.)
-  A2 = _precise_dot(A, A)
-  M2 = _precise_dot(A, E) + _precise_dot(E, A)
-  A4 = _precise_dot(A2, A2)
-  M4 = _precise_dot(A2, M2) + _precise_dot(M2, A2)
-  A6 = _precise_dot(A2, A4)
-  M6 = _precise_dot(A4, M2) + _precise_dot(M4, A2)
-  U = _precise_dot(A, b[7]*A6 + b[5]*A4 + b[3]*A2 + b[1]*ident)
-  V = b[6]*A6 + b[4]*A4 + b[2]*A2 + b[0]*ident
-  Lu = (_precise_dot(A, b[7]*M6 + b[5]*M4 + b[3]*M2) +
-          _precise_dot(E, b[7]*A6 + b[5]*A4 + b[3]*A2 + b[1]*ident))
-  Lv = b[6]*M6 + b[4]*M4 + b[2]*M2
-  return U, V, Lu, Lv, s
-
-@jit
-def _diff_pade9(args):
-  A,E,ident,_ = args
-  s = 0
-  b = (17643225600., 8821612800., 2075673600., 302702400., 30270240., 2162160.,
-       110880., 3960., 90., 1.)
-  A2 = _precise_dot(A, A)
-  M2 = _precise_dot(A, E) + _precise_dot(E, A)
-  A4 = _precise_dot(A2, A2)
-  M4 = _precise_dot(A2, M2) + _precise_dot(M2, A2)
-  A6 = _precise_dot(A2, A4)
-  M6 = _precise_dot(A4, M2) + _precise_dot(M4, A2)
-  A8 = _precise_dot(A4, A4)
-  M8 = _precise_dot(A4, M4) + _precise_dot(M4, A4)
-  U = _precise_dot(A, b[9]*A8 + b[7]*A6 + b[5]*A4 + b[3]*A2 + b[1]*ident)
-  V = b[8]*A8 + b[6]*A6 + b[4]*A4 + b[2]*A2 + b[0]*ident
-  Lu = (_precise_dot(A, b[9]*M8 + b[7]*M6 + b[5]*M4 + b[3]*M2) +
-          _precise_dot(E, b[9]*A8 + b[7]*A6 + b[5]*A4 + b[3]*A2 + b[1]*ident))
-  Lv = b[8]*M8 + b[6]*M6 + b[4]*M4 + b[2]*M2
-  return U, V, Lu, Lv, s
-
-@jit
-def _diff_pade13(args):
-  A,E,ident,A_norm_1 = args
-  s = jnp.maximum(0, jnp.floor_divide(lax.ceil(jnp.log2(A_norm_1 / ell_table_61[13])), 1))
-  two = jnp.array([2.0],A.dtype)
-  A = A * two[0]**-s
-  E = E * two[0]**-s
-  # pade order 13
-  A2 = _precise_dot(A, A)
-  M2 = _precise_dot(A, E) + _precise_dot(E, A)
-  A4 = _precise_dot(A2, A2)
-  M4 = _precise_dot(A2, M2) + _precise_dot(M2, A2)
-  A6 = _precise_dot(A2, A4)
-  M6 = _precise_dot(A4, M2) + _precise_dot(M4, A2)
-  b = (64764752532480000., 32382376266240000., 7771770303897600.,
-       1187353796428800., 129060195264000., 10559470521600.,
-       670442572800., 33522128640., 1323241920., 40840800., 960960.,
-       16380., 182., 1.)
-  W1 = b[13]*A6 + b[11]*A4 + b[9]*A2
-  W2 = b[7]*A6 + b[5]*A4 + b[3]*A2 + b[1]*ident
-  Z1 = b[12]*A6 + b[10]*A4 + b[8]*A2
-  Z2 = b[6]*A6 + b[4]*A4 + b[2]*A2 + b[0]*ident
-  W = _precise_dot(A6, W1) + W2
-  U = _precise_dot(A, W)
-  V = _precise_dot(A6, Z1) + Z2
-  Lw1 = b[13]*M6 + b[11]*M4 + b[9]*M2
-  Lw2 = b[7]*M6 + b[5]*M4 + b[3]*M2
-  Lz1 = b[12]*M6 + b[10]*M4 + b[8]*M2
-  Lz2 = b[6]*M6 + b[4]*M4 + b[2]*M2
-  Lw = _precise_dot(A6, Lw1) + _precise_dot(M6, W1) + Lw2
-  Lu = _precise_dot(A, Lw) + _precise_dot(E, W)
-  Lv = _precise_dot(A6, Lz1) + _precise_dot(M6, Z1) + Lz2
-  return U, V, Lu, Lv, s
-
-@_expm.defjvp
-def _expm_jvp(upper_triangular, primals, tangents):
-  matrix, = primals
-  g, = tangents
-  return expm_frechet(matrix, g, compute_expm=True)
-
 
 @_wraps(scipy.linalg.block_diag)
 @jit
