@@ -111,10 +111,10 @@ def _map_coordinates(input, coordinates, order, mode, cval):
 
 
 @_wraps(scipy.ndimage.map_coordinates, lax_description=textwrap.dedent("""\
-    Only linear interpolation (``order=1``) and modes ``'constant'``,
-    ``'nearest'`` and ``'wrap'`` are currently supported. Note that
-    interpolation near boundaries differs from the scipy function, because we
-    fixed an outstanding bug (https://github.com/scipy/scipy/issues/2640);
+    Only nearest neighbor (``order=0``), linear interpolation (``order=1``) and
+    modes ``'constant'``, ``'nearest'`` and ``'wrap'`` are currently supported.
+    Note that interpolation near boundaries differs from the scipy function,
+    because we fixed an outstanding bug (https://github.com/scipy/scipy/issues/2640);
     this function interprets the ``mode`` argument as documented by SciPy, but
     not as implemented by SciPy.
     """))
