@@ -15,6 +15,7 @@
 
 import functools
 import itertools as it
+import operator
 import types
 
 import numpy as np
@@ -242,3 +243,17 @@ def wrap_name(name, transform_name):
 
 def extend_name_stack(stack, name=''):
   return stack + name + '/'
+
+def canonicalize_axis(axis, num_dims):
+  """Canonicalize an axis in [-num_dims, num_dims) to [0, num_dims)."""
+  axis = operator.index(axis)
+  if not -num_dims <= axis < num_dims:
+    raise ValueError(
+        "axis {} is out of bounds for array of dimension {}".format(
+            axis, num_dims))
+  if axis < 0:
+    axis = axis + num_dims
+  return axis
+
+def ceil_of_ratio(x, y):
+  return -(-x // y)
